#import <Cocoa/Cocoa.h>
#include "SineGenerator.h"
#include <stdio.h>
#include <math.h>

typedef struct {
    @defs(SineGenerator);
} SineGeneratorStruct;

OSStatus audioIOProc (AudioDeviceID		inDevice,
                      const AudioTimeStamp*	inNow,
                      const AudioBufferList*	inInputData,
                      const AudioTimeStamp*	inInputTime,
                      AudioBufferList*		outOutputData,
                      const AudioTimeStamp*	inOutputTime,
                      void*			inClientData)
{
    SineGeneratorStruct* strut = inClientData;
    static	int where = 0;
    UInt32	bufferSize = outOutputData->mBuffers[0].mDataByteSize;
    UInt32	numChannels = outOutputData->mBuffers[0].mNumberChannels;
    UInt32	sampleCount = bufferSize / sizeof(float);
    Float64	sampleRate = strut->outputDescription.mSampleRate;
    UInt32	repeatCount = sampleCount / numChannels;
    float	*buf = (float*) malloc(sampleCount * sizeof(float));
    float	amp = strut->amp, freq = strut->freq;
    int		i, channel;

    for ( i = 0; i < repeatCount; i++ )
        for( channel = 0; channel < numChannels; channel++ )
            buf[2*i+channel] = amp * sin(2 * M_PI * freq / sampleRate * (where+i));
    
    memcpy(outOutputData->mBuffers[0].mData, (void*)buf, bufferSize);

    where += sampleCount/2;
    free(buf);

    return (kAudioHardwareNoError);    
}


@implementation SineGenerator

- (void) setup
{
//    NSLog(@"in Generator:setup");
    OSStatus	err = kAudioHardwareNoError;
    UInt32	propertySize;
    outputDevice = kAudioDeviceUnknown;
    initialized = FALSE;
        
    // get the default output device
    propertySize = sizeof(AudioDeviceID);
    err = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultOutputDevice, &propertySize, (void *) &outputDevice);
    if (err != kAudioHardwareNoError) return;

    // get the size of the buffer
    propertySize = sizeof(UInt32);
    err = AudioDeviceGetProperty(outputDevice, 0, false, kAudioDevicePropertyBufferSize, &propertySize, &outputBufferSize);
    if (err != kAudioHardwareNoError) return;

    // get the description of the audio device
    propertySize = sizeof(AudioStreamBasicDescription);
    err = AudioDeviceGetProperty(outputDevice, 0, false, kAudioDevicePropertyStreamFormat, &propertySize, &outputDescription);
    if (err != kAudioHardwareNoError) return;

    // we want linear PCM i guess
    if (outputDescription.mFormatID != kAudioFormatLinearPCM) return;

    initialized = TRUE;
    
    return;
}

- (void) printConfig
{
//    NSLog(@"in Generator:printConfig");
    printf("sample rate: %f\n", (float)outputDescription.mSampleRate);
    printf("format ID: %u\n", (unsigned int)outputDescription.mFormatID);
    printf("format flags: %i\n", (int)outputDescription.mFormatFlags);
    printf("bytes per packet: %i\n", (int)outputDescription.mBytesPerPacket);
    printf("frames per packet: %i\n", (int)outputDescription.mFramesPerPacket);
    printf("bytes per frame: %i\n", (int)outputDescription.mBytesPerFrame);
    printf("channels per frame: %i\n", (int)outputDescription.mChannelsPerFrame);
    printf("bits per channel: %i\n", (int)outputDescription.mBitsPerChannel);
    
    return;
}

- (void) playSine
{
//    NSLog(@"in Generator:playSine");
    SineGeneratorStruct* def = (SineGeneratorStruct *)self;
    OSStatus err;
    
    if (!initialized) return;
    if (soundPlaying) return;
    
    err = AudioDeviceAddIOProc(outputDevice, audioIOProc, (void *) def);
    if (err != kAudioHardwareNoError) return;
    
    err = AudioDeviceStart(outputDevice, audioIOProc);
    if (err != kAudioHardwareNoError) return;
    
    soundPlaying = TRUE;
    
    return;
}

- (void) stopSine
{
//    NSLog(@"in Generator:stopSine");
    OSStatus err;

    if (!initialized) return;
    if (!soundPlaying) return;
        
    err = AudioDeviceStop(outputDevice, audioIOProc);
    if (err != kAudioHardwareNoError) return;

    err = AudioDeviceRemoveIOProc(outputDevice, audioIOProc);
    if (err != kAudioHardwareNoError) return;
    
    soundPlaying = FALSE;

    return;
}

- (void) setFreq:(float)newFreq
{
//    NSLog(@"frequency: %f", newFreq);
    freq = newFreq;
}

- (void) setAmp:(float)newAmp
{
//    NSLog(@"amplitude: %f", newAmp);
    amp = newAmp;
}

@end