#import "Player.h"

// definitions that control the parameters and function
// of the program
#define SAMPLES_PER_BUFFER (8*1024)
#define multiply 200
#define	kMaxStringSize	1024
#define feederSleepTime 0.2
#define pauseFile 1
#define showPtrs 0
#define preBuffer 0

// the IOProc needs to see these
short		samples[multiply][SAMPLES_PER_BUFFER];
short		missed[SAMPLES_PER_BUFFER];
unsigned int inPt, outPt;
FILE	*file;

static	AudioDeviceID			outputDeviceID;
static	AudioStreamBasicDescription	outputStreamBasicDescription;
void signal_handler(int sig);

/***
name:playAiffFile

purpose:This is the main function that is in charge of setting up the
buffer pointers, registering the IOProc with the audio device, and
filling the buffer with audio data.

input:The path of the file to be played.

output:The audio will be played and information about the file will be
displayed.

functions called:setupOutputDevice, findAudioData, AIFFRead, playAudio, cleanUp
***/

int playAiffFile(char filename[]) {
    static struct sigaction act;
    int error;
    unsigned int bytesRead;

    // trap SIGINT
    act.sa_handler = signal_handler;
    sigfillset(&(act.sa_mask));
    sigaction(SIGINT, &act, NULL);

    // initialize pointers
    inPt = 0; outPt = 0;

    // set up the audio device and check for an error
    if (error = setupOutputDevice() > 0 )
        return error;
    // find the audio data in the sound file
    if (error = findAudioData(filename) > 0 )
        return error;

    // make these global values editable; they function as counters in the program
    int preBufferInt; preBufferInt = preBuffer;
    float feederSleepTimeFloat; feederSleepTimeFloat = feederSleepTime;
    
    // pre-buffer
    if ( preBufferInt == 1 ) {
        NSLog(@"Prebuffering.");
        do {
            // grab data from the file to fill the buffer
            bytesRead = AIFFRead(file, samples[inPt], SAMPLES_PER_BUFFER);
        } while ( (inPt+1)%multiply != 0 );
        NSLog(@"Prebuffering complete.");
    }

    // begin playing the file
    playAudio();
    
    NSLog(@"Playing.");
    NSLog(@"Number of buffers allocated:%i", (int)multiply);

    // play the file while there is still data
    while ( !feof(file) ) {
        // make sure the pointers don't cross
        if ( (((inPt+1)%multiply)!=outPt) && (outPt!=inPt) || preBufferInt == 0) {
            // don't do another prebuffer check
            preBufferInt = 1;
            // grab audio data
            bytesRead = AIFFRead(file, samples[inPt], SAMPLES_PER_BUFFER);
            // fill in zeros if a chunkful is not read
            if ( bytesRead < SAMPLES_PER_BUFFER ) {
                for (; bytesRead < SAMPLES_PER_BUFFER; bytesRead++)
                    samples[inPt][bytesRead] = 0;
            }
            // increase the in pointer
            inPt = ((inPt + 1) % multiply);
            if ( showPtrs ) {
                NSLog(@"inPt:%i\toutPt:%i", inPt, outPt);
            }
        }
        // if the file is paused, wait to read later
        else if ( pauseFile ) {
            NSLog(@"File is \"paused\". Waiting %.2f seconds.", (float)feederSleepTime);
            usleep ( feederSleepTime * 1000 * 1000);
        }
        // if simulating losing data, discard the audio data
        // read from the file
        else if ( !pauseFile ){
            NSLog(@"Data was lost because buffers are full.");
            NSLog(@"Waiting %.2f seconds.", feederSleepTimeFloat);
            bytesRead = AIFFRead(file, missed, SAMPLES_PER_BUFFER);
            usleep ( feederSleepTime * 1000 * 1000);
        }
    }

    NSLog(@"Cleaning up.");
    // clean up before exiting
    cleanUp();
    NSLog(@"Finished cleaning up.");
    NSLog(@"Goodbye.");

    return 0;
}

/***
name:audioDeviceIOProc

purpose:This function is registered with the audio device; it takes
data from the sample buffer and places it in the audio device's
buffer.

input:An audio device, which will provide buffers and audio device
information.

output: The sound from the audio file.

functions called:none
***/
OSStatus audioDeviceIOProc(AudioDeviceID inDevice,
                           const AudioTimeStamp *inNow,
                           const AudioBufferList *inInputData,
                           const AudioTimeStamp *inInputTime,
                           AudioBufferList *outOutputData,
                           const AudioTimeStamp *inOutputTime,
                           void *inClientData)
{
    unsigned int sampleIndex;
    float *outBuffer;
    float scale = (1.0f / SHRT_MAX);    

    // make sure our device doesn't change during playback
    assert(outOutputData->mNumberBuffers == 1);
    assert(outOutputData->mBuffers[0].mNumberChannels == 2);
    assert(outOutputData->mBuffers[0].mDataByteSize == (SAMPLES_PER_BUFFER * sizeof(float)));

    // find the buffer that we need to fill
    outBuffer = (float *)outOutputData->mBuffers[0].mData;

    // make sure there is audio data to play by checking the
    // buffer pointers don't cross
    if ( ((outPt + 1) % multiply) != inPt) {
        // convert the audio data into floats and put it in
        // the buffer
        for (sampleIndex = 0; sampleIndex < SAMPLES_PER_BUFFER; sampleIndex++) {
            outBuffer[sampleIndex] = samples[outPt][sampleIndex] * scale;
        }
        // increase the pointer
        outPt = ((outPt + 1) % multiply);
    }
    // if there's no data to play, play nothing
    else {
        for (; sampleIndex < SAMPLES_PER_BUFFER; sampleIndex++)
            outBuffer[sampleIndex] = sampleIndex;
    }
    
    return (kAudioHardwareNoError);
}

/***
name:AIFFRead

purpose:Read data from the audio file and 

input: the audio file pointer, a sample buffer, how many samples to read

output: returns the number of samples read

functions called: fread
***/
unsigned int AIFFRead(FILE *infile, short samples[], unsigned int sampleCount)
{
    unsigned int index;

    // read from the file and record how many samples were read
    sampleCount = fread(samples, sizeof(short), sampleCount, infile);

    // CD data is stored in little-endian order, but we want big endian, being on PPC.
    for (index = 0; index < sampleCount; index++) {
        short sample;

        sample = samples[index];
        sample = ((sample & 0xff00) >> 8) | ((sample & 0x00ff) << 8);
        samples[index] = sample;
    }

    return sampleCount;
}

/***
name:findAudioData

purpose:Find the audio data in an AIFC file.

input:a file path, a file pointer

output:a file pointer that is set to the audio data

functions called: none
***/
int findAudioData(char filename[]) {
    NSString *path = [[NSString alloc] initWithCString:(const char *)filename];
    const char *pathStr;
    AIFFChunkHeader chunkHeader;
    AIFFGenericChunk chunk;
    AIFFSSNDData ssndData;

    // make sure file path is correct
    pathStr = [path fileSystemRepresentation];
    file = fopen(pathStr, "r");
    if (!file) {
        perror(pathStr);
        return 21;
    }

    // check the for the FORM header
    fread(&chunkHeader, 1, sizeof(chunkHeader), file);
    if (chunkHeader.chunkID != 'FORM') {
        fprintf(stderr, "chunkID is not 'FORM'\n");
        fclose(file);
        return 20;
    }

    // check the file type
    if (chunkHeader.fileType != 'AIFC') {
        fprintf(stderr, "file format is not 'AIFC'\n");
        fclose(file);
        return 22;
    }

    // Skip up to the 'SSND' chunk
    while (1) {
        // read a chunk
        fread(&chunk, 1, sizeof(chunk), file);
        
        //check if it's the SSND chunk
        if (chunk.chunkID == 'SSND')
            // exit the loop if it is
            break;

        // show what chunk is being skipped
        fprintf(stderr, "skipping chunk %c%c%c%c\n", (chunk.chunkID >> 24) & 0xff, (chunk.chunkID >> 16) & 0xff, (chunk.chunkID >> 8) & 0xff, (chunk.chunkID >> 0) & 0xff);

        // skip over that chunk
        fseek(file, chunk.chunkSize, SEEK_CUR);
    }

    // find the sound chunk and display its attributes
    fprintf(stderr, "Found SSND, size = %d\n", chunk.chunkSize);
    fread(&ssndData, 1, sizeof(ssndData), file);
    fprintf(stderr, "offset = %d\n", ssndData.offset);
    fprintf(stderr, "blockSize = %d\n", ssndData.blockSize);
    
    return 0;
}

/***
name:playAudio

purpose:Register the IOProc with an audio device and begin playback.

input:the default audio device, IOProc

output:errors

functions called:AudioDeviceAddIOProc, AudioDeviceStart
***/
int playAudio(){
    OSStatus status;

    // register the IOProc with the audio device
    status = AudioDeviceAddIOProc(outputDeviceID, audioDeviceIOProc, NULL);
    if (status) {
        NSLog(@"AudioDeviceAddIOProc: returned %d", status);
        return 15;
    }

    // tell the audio device to begin playback
    status = AudioDeviceStart(outputDeviceID, audioDeviceIOProc);
    if (status) {
        NSLog(@"AudioDeviceStart: returned %d", status);
        return 16;
    }
    
    return 0;
}

/***
name:stopAudio

purpose:Remove the IOProc from the audio device and stop playback.

input:the default audio device, IOProc

output:errors

functions called:AudioDeviceRemoveIOProc, AudioDeviceStop
***/
int stopAudio() {
    OSStatus status;

    // tell the audio device to stop playback
    status = AudioDeviceStop(outputDeviceID, audioDeviceIOProc);
    if (status) {
        NSLog(@"AudioDeviceStop: returned %d", status);
        return 16;
    }
    
    // remove the IOProc with the audio device
    status = AudioDeviceRemoveIOProc(outputDeviceID, audioDeviceIOProc);
    if (status) {
        NSLog(@"AudioDeviceRemoveIOProc: returned %d", status);
        return 15;
    }
    
    return 0;
}

/***
name:cleanUp

purpose:Stop audio and close file.

input:none

output:none

functions called:stopAudio
***/
int cleanUp() {

    // stop output
    stopAudio();
    // close audio file
    fclose(file);
    
    return 0;
}

/***
name:setupOutputDevice

purpose:Find the default audio device and initialize it.

input:none

output:an initialized audio device

functions called: HAL routines
***/
int setupOutputDevice () {
    OSStatus status;
    UInt32 propertySize, bufferByteCount, theUInt32;
    char stringie[kMaxStringSize];

    NSLog(@"***Device Information***\n");
    // find the default audio device
    propertySize = sizeof(outputDeviceID);
    status = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultOutputDevice, &propertySize, &outputDeviceID);
    if (status) {
        NSLog(@"AudioHardwareGetProperty returned %d", status);
        return 10;
    }

    //checkt that there is a device
    if (outputDeviceID == kAudioDeviceUnknown) {
        NSLog(@"AudioHardwareGetProperty: outputDeviceID is kAudioDeviceUnknown");
        return 11;
    }

    // get the device name
    propertySize = kMaxStringSize;
    status = AudioDeviceGetProperty(outputDeviceID, 0, 0, kAudioDevicePropertyDeviceName, &propertySize, stringie);
    if(status == 0)
        NSLog(@"Device Name:%s", stringie);
    else
        NSLog(@"Couldn't get the device name.");

    // get the manufacturer name
    propertySize = kMaxStringSize;
    status = AudioDeviceGetProperty(outputDeviceID, 0, 0, kAudioDevicePropertyDeviceManufacturer, &propertySize, stringie);
    if(status == 0)
        NSLog(@"Device Manufacturer:%s", stringie);
    else
        NSLog(@"Couldn't get the manufacturer name.");

    // check if something is in the headphone jack; just for fun
    if(AudioDeviceGetPropertyInfo(outputDeviceID, 0, 0, kAudioDevicePropertyJackIsConnected, NULL, NULL) == 0)
    {
        theUInt32 = 0;
        propertySize = sizeof(UInt32);
        status = AudioDeviceGetProperty(outputDeviceID, 0, 0, kAudioDevicePropertyJackIsConnected, &propertySize, &theUInt32);
        if(status == 0)
        {
            if (theUInt32)
                NSLog(@"Output jack connected.");
            else
                NSLog(@"Output jack not connected.");
        }
        else
            NSLog(@"Couldn't get the whether or not the output jack is connected.");
    }

    // configure the output device
    propertySize = sizeof(bufferByteCount);
    bufferByteCount = SAMPLES_PER_BUFFER * sizeof(float);
    status = AudioDeviceSetProperty(outputDeviceID, NULL, 0, NO, kAudioDevicePropertyBufferSize, propertySize, &bufferByteCount);
    if (status) {
        NSLog(@"AudioDeviceSetProperty: returned %d when setting kAudioDevicePropertyBufferSize to %d", status, SAMPLES_PER_BUFFER);
        return 12;
    }

    // get the output device description
    propertySize = sizeof(outputStreamBasicDescription);
    status = AudioDeviceGetProperty(outputDeviceID, 0, NO, kAudioDevicePropertyStreamFormat, &propertySize, &outputStreamBasicDescription);
    if (status) {
        NSLog(@"AudioDeviceGetProperty: returned %d when getting kAudioDevicePropertyStreamFormat", status);
        return 13;
    }

    // print the device's description
    NSLog(@"***Hardware Format***\n");
    NSLog(@"Sample Rate:%f", outputStreamBasicDescription.mSampleRate);
    NSLog(@"FormatID:%c%c%c%c",
          (outputStreamBasicDescription.mFormatID & 0xff000000) >> 24,
          (outputStreamBasicDescription.mFormatID & 0x00ff0000) >> 16,
          (outputStreamBasicDescription.mFormatID & 0x0000ff00) >>  8,
          (outputStreamBasicDescription.mFormatID & 0x000000ff) >>  0);
    NSLog(@"Bytes Per Frame:%d", outputStreamBasicDescription.mBytesPerFrame);
    NSLog(@"Channels Per Frame:%d", outputStreamBasicDescription.mChannelsPerFrame);
    NSLog(@"Bits Per Channel:%d", outputStreamBasicDescription.mBitsPerChannel);

    if (outputStreamBasicDescription.mFormatID != kAudioFormatLinearPCM) {
        NSLog(@"Default Audio Device doesn't support Linear PCM.");
        return 14;
    }

    return 0;
}

/***
name:signal_handler

purpose:Trap the SIGINT signal, which is sent when the user enters ^C.

input:signal

output:notification of termination

functions called:cleanUp
***/
void signal_handler(int sig) {
    cleanUp();
    printf("\nGoodbye.\n");
    exit(0);
}